//
// Cx.c -- Functions related to setting up a hotkey with the commodity system.
//

/**/
/*                            I N C L U D E S									*/
/**/
#include <exec/types.h>								// For UBYTE etc.
#include <clib/alib_protos.h>						// Protos for ArgArrayInit() etc.
#include <clib/commodities_protos.h>			// For CxBroker() etc.
#include <clib/exec_protos.h>						// Protos for CreateMsgPort() etc.
#include <clib/intuition_protos.h>				// Protos for ScreenToFront() etc.
#include <libraries/commodities.h>				// For Cx structs etc.
#include <pragmas/commodities_pragmas.h>		// Pragmas for commodities.library.
#include <pragmas/exec_pragmas.h>				// Pragmas for exec.library.
#include <pragmas/intuition_pragmas.h>			// Pragmas for intuition.library.

#include "/include/Prefs.h"						// For struct WPrefs.
#include "/include/Catalog.h"						// V37 compatible locale.
#define CATCOMP_NUMBERS
#include "/include/Watcher_strings.h"			// All strings used in Watcher.

/**/
/*                             I M P O R T S										*/
/**/
/*------------*/
/* Variables: */
/*------------*/
extern struct Library		*CxBase;				// commodities.library.
extern struct Library		*IconBase;			// Used by ArgArrayInit() etc.
extern struct Library		*IntuitionBase;	// intuition.library.
extern struct MsgPort		*CxMsgPort;			// Message port for getting input.
extern CxObj					*CxMyBroker;		// Our Cx broker.
extern CxObj					*CxFilterObj;		// Filter object.
extern CxObj					*CxSenderObj;		// Sender object.
extern CxObj					*CxTranslateObj;	// Translator object.
extern UBYTE					**tt;					// Tooltypes.
extern UBYTE					*Hotkey;				// Hotkey for commodity.
extern struct Screen			*PubScr;				// Public screen pointer.
extern struct Window			*Win;					// Watcher window.
extern struct Window			*SLBWin;				// SLB window.
extern BOOL						AppInfoActive;		// Determines if App Info is launched.
extern struct AppIcon		*ai;					// Applicaton Icon.
extern struct WPrefs			*WPrefs;				// Preferences for Watcher.
extern struct MsgPort		*amp;					// Message port for App Info.
extern struct DiskObject	*dobj;				// Disk Object for App Info.
extern BOOL						HideMode;			// CX_POPUP.

/*------------*/
/* Functions: */
/*------------*/
BOOL Inform					(struct Window *w, STRPTR Title, STRPTR Text, APTR args, ...);
void DeAllocateWindow	(void);
void AllocateWindow		(struct Screen *pubScr);

/**/
/*                               CxSetup()										*/
/**/
BOOL CxSetup (int argc, char *argv [])
{
	/*-------------*/
	/* L O C A L S */
	/*-------------*/
	struct NewBroker	newbroker =
	{
		NB_VERSION,										// Version of this structure.
		"Watcher Coolness",							// Name of this commodity.
		"Watcher V1.0 1993 Franz Hemmer",		// Title of commodity.
		"The Ultimate Watcher Utility",			// Commodity description.
		NBU_UNIQUE | NBU_NOTIFY,					// Unique = Yes!
		COF_SHOW_HIDE,									// Window = Yes!
		0,													// Priority.
		0,													// MsgPort Cx talks to. (Later!)
		0													// Reserved.
	};

	/*---------*/
	/* C O D E */
	/*---------*/
	/*---------------------*/
	/* Localize newbroker. */
	/*---------------------*/
	newbroker.nb_Name		= XGetString (M_COMMODITY_NAME);
	newbroker.nb_Title	= XGetString (M_COMMODITY_TITLE);
	newbroker.nb_Descr	= XGetString (M_COMMODITY_DESCRIPTION);

	/*-----------------------------------------------*/
	/* First we must get the CX_xxx tooltype values. */
	/*-----------------------------------------------*/
	if (!(tt = ArgArrayInit (argc, argv)))
	{
		goto panic;
	}

	/*----------------------*/
	/* Check the tooltypes. */
	/*----------------------*/
	newbroker.nb_Pri = (BYTE) ArgInt (tt, "CX_PRIORITY", 0);
	Hotkey = ArgString (tt, "CX_POPKEY", "rawkey control esc");

	/*-----------------------------*/
	/* Create message port for Cx. */
	/*-----------------------------*/
	if (!(CxMsgPort = CreateMsgPort ()))
	{
		goto panic;
	}

	/*--------------------------------------------------*/
	/* Setup message port into the newbroker structure. */
	/*--------------------------------------------------*/
	newbroker.nb_Port = CxMsgPort;

	/*---------------------------------------*/
	/* Create our broker with the CX system. */
	/*---------------------------------------*/
	if (!(CxMyBroker = CxBroker (&newbroker, NULL)))
	{
		goto panic;
	}

	/*-----------------------*/
	/* Create filter object. */
	/*-----------------------*/
	if (!(CxFilterObj = CxFilter (Hotkey)))
	{
		goto panic;
	}

	/*----------------------------------*/
	/* Add filter object to our broker. */
	/*----------------------------------*/
	AttachCxObj (CxMyBroker, CxFilterObj);

	/*--------------------------------------------------------------------------*/
	/* Create sender object. This object will send message to our message port. */
	/*--------------------------------------------------------------------------*/
	if (!(CxSenderObj = CxSender (CxMsgPort, 1L)))
	{
		goto panic;
	}

	/*--------------------------------------*/
	/* Add the sender object to our broker. */
	/*--------------------------------------*/
	AttachCxObj (CxFilterObj, CxSenderObj);

	/*------------------------------------------------------------------------------*/
	/* Create tanslator object. Passing NULL to CxTranlate() indicates that we want */
	/* the Input Event to stop at our broker.													  */
	/*------------------------------------------------------------------------------*/
	if (!(CxTranslateObj = CxTranslate (NULL)))
	{
		goto panic;
	}

	/*--------------------------------------------------------*/
	/* Add the translator (terminator in fact) to our broker. */
	/*--------------------------------------------------------*/
	AttachCxObj (CxFilterObj, CxTranslateObj);

	/*---------------------------------------------------------*/
	/* Check the accumulated errors for our CxMyBroker object. */
	/*---------------------------------------------------------*/
	if (CxObjError (CxFilterObj))
	{
		goto panic;
	}

	/*----------------------------------------*/
	/* Now we can finally turn ON the broker: */
	/*----------------------------------------*/
	ActivateCxObj (CxMyBroker, 1L);

	/*------------------------------------*/
	/* Don't need the tooltypes any more. */
	/*------------------------------------*/
	ArgArrayDone ();

	return (TRUE);

 panic:

	if (CxMyBroker)
	{
		DeleteCxObjAll (CxMyBroker);
	}
	if (CxMsgPort)
	{
		DeleteMsgPort (CxMsgPort);
	}
	if (tt)
	{
		ArgArrayDone ();
	}
	return (FALSE);
}

/**/
/*                             CxShutDown()										*/
/**/
void CxShutDown (void)
{
	/*-------------*/
	/* L O C A L S */
	/*-------------*/
	struct CxMsg *cmsg;

	/*---------*/
	/* C O D E */
	/*---------*/
	/*---------------------------------------*/
	/* Remove broker and related Cx objects. */
	/*---------------------------------------*/
	if (CxMyBroker)
	{
		DeleteCxObjAll (CxMyBroker);
	}

	if (CxMsgPort)
	{
		/*--------------------------------------------*/
		/* Remove all pending messages from the port. */
		/*--------------------------------------------*/
		while (cmsg = (struct CxMsg *) GetMsg (CxMsgPort))
		{
			ReplyMsg ((struct Message *) cmsg);
		}
	}

	/*--------------------------*/
	/* Delete the message port. */
	/*--------------------------*/
	if (CxMsgPort)
	{
		DeleteMsgPort (CxMsgPort);
	}
}

/**/
/*                            HandleCxIDCMP()									*/
/**/
BOOL HandleCxIDCMP (ULONG msgid, ULONG msgtype)
{
	/*-------------*/
	/* L O C A L S */
	/*-------------*/
	BOOL rc = FALSE;

	/*---------*/
	/* C O D E */
	/*---------*/
	switch (msgtype)
	{
		case CXM_IEVENT :
		{
			switch (msgid)
			{
				case 1 :
				{
					/*----------------------*/
					/* Activated by hotkey! */
					/*----------------------*/
					if (!Win)
					{
						/*----------------------*/
						/* Open Watcher window. */
						/*----------------------*/
						HideMode = FALSE;
						AllocateWindow (PubScr);

						/*-------------------------------------*/
						/* Check if we should launch App Info. */
						/*-------------------------------------*/
						if ((WPrefs->wp_AppInfoAutoStart) && (!AppInfoActive))
						{
							/*----------------------------------*/
							/* Add the application icon to the. */
							/*----------------------------------*/
							if (!(ai = AddAppIconA (1, 0, "App Info", amp, NULL, dobj, NULL)))
							{
								Inform (Win, "Error:", "%s", XGetString (M_CANT_LAUNCH_APPINFO));
							}
							else
							{
								AppInfoActive = TRUE;
							}
						}
					}
					ActivateWindow (Win);
					ScreenToFront (PubScr);
					break;
				}
				default :
				{
					break;
				}
			}
			break;
		}
		case CXM_COMMAND :
		{
			switch (msgid)
			{
				case CXCMD_DISABLE :
				case CXCMD_DISAPPEAR :
				{
					/*------------------------------------------*/
					/* Check that the SLB window is not active. */
					/*------------------------------------------*/
					if (!SLBWin)
					{
						/*---------------------------------------------*/
						/* Check if wee need to close our window down. */
						/*---------------------------------------------*/
						if (Win)
						{
							/*--------------------------------------*/
							/* Check if we need to remove App Info. */
							/*--------------------------------------*/
							if (AppInfoActive && ai)
							{
								/*---------------------*/
								/* We remove App Info. */
								/*---------------------*/
								RemoveAppIcon (ai);
								ai = NULL;
								AppInfoActive = FALSE;
							}
							/*-------------------------------*/
							/* We remove our Watcher window. */
							/*-------------------------------*/
							DeAllocateWindow ();
						}
						else
						{
							/*---------------------*/
							/* Window not present. */
							/*---------------------*/
							Inform (NULL, "Watcher:", "%s", XGetString (M_ALREADY_DISABLED));
						}
					}
					else
					{
						/*----------------------------------------------------*/
						/* SLB window active which needs to be removed first. */
						/*----------------------------------------------------*/
						UBYTE *z = XGetString (M_WARNING);
						Inform (NULL, z, "%s", XGetString (M_QUIT_SLB_FIRST));
					}
					break;
				}
				case CXCMD_ENABLE :
				case CXCMD_APPEAR :
				{
					/*----------------------------------------*/
					/* Check if the window is already active. */
					/*----------------------------------------*/
					if (Win)
					{
						Inform (NULL, "Watcher:", "%s", XGetString (M_ALREADY_ENABLED));
					}
					else
					{
						/*-------------------------------*/
						/* We enable the user interface. */
						/*-------------------------------*/
						AllocateWindow (PubScr);
						if ((WPrefs->wp_AppInfoAutoStart) && (!AppInfoActive))
						{
							/*----------------------------------*/
							/* Add the application icon to the. */
							/*----------------------------------*/
							if (!(ai = AddAppIconA (1, 0, "App Info", amp, NULL, dobj, NULL)))
							{
								Inform (Win, "Error:", "%s", XGetString (M_CANT_LAUNCH_APPINFO));
							}
							else
							{
								AppInfoActive = TRUE;
							}
						}

						/*--------------------------------------------------------------*/
						/* And we put the screen where Watcher is located to the front. */
						/*--------------------------------------------------------------*/
						ScreenToFront (PubScr);
						HideMode = FALSE;
					}
					break;
				}
				case CXCMD_KILL :
				{
					/*---------------*/
					/* Kill request. */
					/*---------------*/
					rc = TRUE;
					break;
				}
				case CXCMD_UNIQUE :
				{
					UBYTE *z = XGetString (M_INFORMATION);
					Inform (NULL, z, "%s", XGetString (M_CX_UNIQUE));
					break;
				}
			}
			break;
		}
		default :
		{
			break;
		}
	}
	return (rc);
}
