//
// datatype.c -- Data type related functions.
//
/**/
/*                            I N C L U D E S									*/
/**/
#include <exec/types.h>								// For UBYTE etc.
#include <exec/memory.h>							// For MEMF_ANY etc.
#include <clib/alib_protos.h>						// Protos for DoMethod() etc.
#include <clib/datatypes_protos.h>				// Protos for ObtainDataTypeA() etc.
#include <clib/dos_protos.h>						// Protos for Lock() etc.
#include <clib/exec_protos.h>						// Protos for OpenLibrary() etc.
#include <clib/graphics_protos.h>				// Protos for graphics funcs.
#include <clib/iffparse_protos.h>				// Protos for IDtoStr() etc.
#include <clib/intuition_protos.h>				// Protos for OpenScreen() etc.
#include <datatypes/datatypes.h>					// For struct DataType etc.
#include <datatypes/datatypesclass.h>			// For DTST_FILE etc.
#include <datatypes/pictureclass.h>				// For PDTA_Remap etc.
#include <graphics/gfx.h>							// For struct BitMap.
#include <intuition/intuitionbase.h>			// For struct IntuitionBase.
#include <pragmas/datatypes_pragmas.h>			// Pragmas for datatypes.library.
#include <pragmas/dos_pragmas.h>					// Pragmas for dos.library.
#include <pragmas/exec_pragmas.h>				// Pragmas for exec.library.
#include <pragmas/iffparse_pragmas.h>			// Pragmas for iffparse.library.
#include <pragmas/graphics_pragmas.h>			// Pragmas for graphics.library.
#include <pragmas/intuition_pragmas.h>			// Pragmas for intuition.library.
#include <stdlib.h>									// Proto for exit().
#include <stdio.h>									// Protos for printf() etc.
#include <string.h>									// Protos for strcpy() etc.

/**/
/*                             I M P O R T S										*/
/**/
/*------------*/
/* Variables: */
/*------------*/
extern struct ExecBase		*SysBase;				// supplied by startup module.
extern struct DosLibrary	*DOSBase;				// do.
extern struct Library		*DataTypesBase;		// datatypes.library.
extern struct Library		*GfxBase;				// graphics.library.
extern struct Library		*IFFParseBase;			// iffparse.library.
extern struct Library		*IntuitionBase;		// intuition.library.

/**/
/*                          GetDataTypeString()									*/
/**/
void GetDataTypeString (struct DataType *dt, UBYTE *strbuf)
{
	/*-----------------------------------*/
	/* Put data type ID in users buffer. */
	/*-----------------------------------*/
	strcpy (strbuf, dt->dtn_Header->dth_Name);
	strcat (strbuf, " ");
	strcat (strbuf, GetDTString (dt->dtn_Header->dth_GroupID));
}

/**/
/*                           GetPictureInfo()									*/
/**/
Object *GetPictureInfo (struct DataType *dt, BPTR fileLock, UBYTE *infoBuf)
{

	/*-------------*/
	/* L O C A L S */
	/*-------------*/
	UBYTE					fn [256];
	struct dtFrameBox	dtf	= { NULL };
	struct FrameInfo	fri	= { NULL };
	Object				*o		= NULL;

	/*---------*/
	/* C O D E */
	/*---------*/
	/*----------------*/
	/* Get data type. */
	/*----------------*/
	if (dt->dtn_Header->dth_GroupID == GID_PICTURE)
	{
		/*----------------*/
		/* Get file name. */
		/*----------------*/
		if (NameFromLock (fileLock, fn, 256))
		{
			/*------------------------*/
			/* Create picture object. */
			/*------------------------*/
			o = NewDTObject ((APTR) fn,
								  DTA_SourceType,	DTST_FILE,
								  DTA_GroupID,		GID_PICTURE,
								  PDTA_Remap,		FALSE,
								  TAG_DONE);

			/*-----------------------------------------------*/
			/* Ask object what kind of environment it needs. */
			/*-----------------------------------------------*/
			dtf.MethodID				= DTM_FRAMEBOX;
			dtf.dtf_FrameInfo			= &fri;
			dtf.dtf_ContentsInfo		= &fri;
			dtf.dtf_SizeFrameInfo	= sizeof (struct FrameInfo);
			if (DoMethodA (o, (Msg) &dtf) && fri.fri_Dimensions.Depth)
			{
				sprintf (infoBuf, "%ld x %ld x %ld",
							(ULONG) fri.fri_Dimensions.Width,
							(ULONG) fri.fri_Dimensions.Height,
							(ULONG) fri.fri_Dimensions.Depth);
			}
		}
	}
	return (o);
}

/**/
/*                           DisplayPicture()									*/
/**/
void DisplayPicture (struct DataType *dt, Object *o)
{
	/*-------------*/
	/* L O C A L S */
	/*-------------*/
	struct dtFrameBox		dtf		= { NULL };
	struct FrameInfo		fri		= { NULL };
	struct gpLayout		gpl;
	ULONG						modeid	= INVALID_ID;
	ULONG						*cregs	= NULL, numcolors, i, r, g, b;
	struct BitMapHeader	*bmhd;
	struct Screen			*scr;
	struct BitMap			*bm;
	struct Message			*msg = NULL;
	ULONG						signal;
	struct Window			*w;

	/*---------*/
	/* C O D E */
	/*---------*/
	/*-----------------------------*/
	/* Act according to data type. */
	/*-----------------------------*/
	switch (dt->dtn_Header->dth_GroupID)
	{
		case GID_PICTURE :
		{
			/*-----------------------------------------------*/
			/* Ask object what kind of environment it needs. */
			/*-----------------------------------------------*/
			dtf.MethodID				= DTM_FRAMEBOX;
			dtf.dtf_FrameInfo			= &fri;
			dtf.dtf_ContentsInfo		= &fri;
			dtf.dtf_SizeFrameInfo	= sizeof (struct FrameInfo);
			if (DoMethodA (o, (Msg) &dtf) && fri.fri_Dimensions.Depth)
			{
				/*-----------------------------------*/
				/* Layout the object on our process. */
				/*-----------------------------------*/
				gpl.MethodID	= DTM_PROCLAYOUT;
				gpl.gpl_GInfo	= NULL;
				gpl.gpl_Initial= 1;
				DoMethodA (o, (Msg) &gpl);

				/*------------------*/
				/* Get object info. */
				/*------------------*/
				GetDTAttrs (o,
								PDTA_ModeID,			&modeid,
								PDTA_CRegs,				&cregs,
								PDTA_NumColors,		&numcolors,
								PDTA_BitMapHeader,	&bmhd,
								PDTA_BitMap,			&bm,
								TAG_DONE);

				/*-----------------------------*/
				/* Make sure we have a bitmap. */
				/*-----------------------------*/
				if (bm)
				{
					/*--------------*/
					/* Open screen. */
					/*--------------*/
					scr = OpenScreenTags (NULL,
												SA_Width,			(ULONG) fri.fri_Dimensions.Width,
												SA_Height,			(ULONG) fri.fri_Dimensions.Height,
												SA_Depth,			(ULONG) fri.fri_Dimensions.Depth,
												SA_DisplayID,		modeid,
												SA_Interleaved,	TRUE,
												SA_ShowTitle,		FALSE,
												SA_Quiet,			TRUE,
												SA_Behind,			TRUE,
												SA_BitMap,			bm,
												SA_AutoScroll,		TRUE,
												TAG_DONE);
					if (cregs)
					{
						numcolors = 2 << (scr->RastPort.BitMap->Depth - 1);
						for (i = 0; i < numcolors; i++)
						{
							r = cregs[i * 3 + 0];
							g = cregs[i * 3 + 1];
							b = cregs[i * 3 + 2];
							SetRGB32 (&scr->ViewPort, i, r, g, b);
						}
					}

					/*-----------------------------------*/
					/* Bring the screen to the front now */
					/*-----------------------------------*/
					ScreenToFront (scr);

					if (w = OpenWindowTags (NULL,
													WA_Left,				0,
													WA_Top,				0,
													WA_Width,			1,
													WA_Height,			1,
													WA_IDCMP,			IDCMP_MOUSEBUTTONS	|
																			IDCMP_RAWKEY			|
																			IDCMP_INACTIVEWINDOW,
													WA_CustomScreen,	scr,
													WA_Backdrop,		TRUE,
													WA_Borderless,		TRUE,
													WA_RMBTrap,			TRUE,
													WA_Activate,		TRUE,
													TAG_END))
					{
						signal = Wait (1L << w->UserPort->mp_SigBit);
						while (msg = GetMsg (w->UserPort))
						{
							ReplyMsg (msg);
						}
						CloseWindow (w);
					}
					CloseScreen (scr);
				}
			}
			break;
		}
	}
}
