//
// Lists.c -- Functions for handling the SLB lists.
//
/**/
/*                            I N C L U D E S									*/
/**/
#include <exec/types.h>								// For UBYTE etc.
#include <exec/memory.h>							// For MEMF_ANY etc.
#include <exec/execbase.h>							// For struct ExecBase.
#include <clib/alib_protos.h>						// Protos for NewList() etc.
#include <clib/diskfont_protos.h>				// Protos for AvailFonts
#include <clib/dos_protos.h>						// Protos for Lock() etc.
#include <clib/exec_protos.h>						// Protos for FindTask() etc.
#include <clib/gadtools_protos.h>				// Protos for CreateContext() etc.
#include <clib/graphics_protos.h>				// Protos for SetRast() etc.
#include <clib/intuition_protos.h>				// Protos for CloseScreen() etc.
#include <clib/utility_protos.h>					// Protos for Stricmp() etc.
#include <ctype.h>									// For isalnum() etc.
#include <libraries/gadtools.h>					// For struct NewGadget etc.
#include <pragmas/diskfont_pragmas.h>			// Pragmas for diskfont.library.
#include <pragmas/dos_pragmas.h>					// Pragmas for dos.library.
#include <pragmas/exec_pragmas.h>				// Pragmas for exec.library.
#include <pragmas/gadtools_pragmas.h>			// Pragmas for gadtools.library.
#include <pragmas/graphics_pragmas.h>			// Pragmas for graphics.library.
#include <pragmas/intuition_pragmas.h>			// Pragmas for intuition.library.
#include <pragmas/utility_pragmas.h>			// Pragmas for utility.library.
#include <stdio.h>									// For BUFSIZ.
#include <stdlib.h>									// Proto for atoi() etc.
#include <string.h>									// Proto for strcat() etc.

#include "/include/Catalog.h"						// V37 compatible locale.
#define CATCOMP_NUMBERS
#include "/include/Watcher_strings.h"			// All strings used in Watcher.

#include "/include/Nodes.h"						// My node structures.

/**/
/*                             I M P O R T S										*/
/**/
/*------------*/
/* Variables: */
/*------------*/
extern struct ExecBase			*SysBase;		// supplied by startup module.
extern struct DosLibrary		*DOSBase;		// do.
extern struct Library			*DiskfontBase;	// diskfont.library.
extern struct Library			*GadToolsBase;	// gadtools.library.
extern struct Library			*GfxBase;		// graphics.library.
extern struct IntuitionBase	*IntuitionBase;// intuition.library.
extern struct Library			*UtilityBase;	// utility.library.

extern APTR							mem_pool;
extern BOOL							mem_pool_ok;
extern struct List				*MyScrList;
extern struct List				*MyLibList;
extern struct List				*MyDevList;
extern struct List				*MyTskList;
extern struct List				*MyPrtList;
extern struct List				*MyResList;
extern struct List				*MyFntList;
extern struct List				*MyVolList;
extern struct List				*MyAssList;

extern struct TextAttr			topaz8;			// Hardcoded font for SLB (so far).
extern struct Screen				*PubScr;			// Public screen.
extern struct Window				*SLBWin;			// System List Browser window.
extern struct Gadget				*SLBGList;		// List of gadgets.
extern struct Menu				*SLBMenus;		// Menu for SLB.
extern struct Process			*p;				// Returned by FindTask().
extern struct Window				*Win;				// Watcher window.
extern struct Window				*WindowPtr;		// Our cli window.
extern struct Gadget				*SLBGads [GADMAX];// Gadget pointer array.
extern struct VisualInfo		*VisualInfo;	// For menu.
extern struct Requester			SleepReq;
extern struct Requester			SleepReq2;
extern UBYTE						*CycleText [];
extern UWORD						DisplayType;

/**/
/*                              P R O T O S										*/
/**/
void				CleanExit					(WORD ExitCode, STRPTR ExitText, ULONG DosExitCode);

void				AllocSLBWindow				(void);
void				FreeSLBWindow				(void);
void				CreateSLBGadgets			(void);
void				FreeSLBGadgets				(void);
void				CreateSLBMenu				(void);
void				FreeSLBMenu					(void);

BOOL				AllocAllLists				(void);
void				ResetAllLists				(void);
void				RemAllNodes					(void);
void				FreeAllLists				(void);
void				CollectAllLists			(void);
void				CollectScrList				(void);
void				CollectLibList				(void);
void				CollectDevList				(void);
void				CollectTskList				(void);
void				CollectPrtList				(void);
void				CollectResList				(void);
void				CollectFntList				(void);
void				CollectVolList				(void);
void				UpdateScrList				(BOOL Full);
void				UpdateLibList				(BOOL Full);
void				UpdateDevList				(BOOL Full);
void				UpdateTskList				(BOOL Full);
void				UpdatePrtList				(BOOL Full);
void				UpdateResList				(BOOL Full);
void				UpdateFntList				(BOOL Full);
void				UpdateVolList				(BOOL Full);
void				UpdateAssList				(BOOL Full);
void				CreateANode					(struct List *l, struct Node *n);
void				FreeAllNodes				(struct List *l);
struct Node 	*GetNodeNum					(struct List *l, ULONG num);

BOOL				SleepWindow					(struct Window *w, struct Requester *r);
void				WakeUpWindow				(struct Window *w, struct Requester *r);

void				SortList						(struct List *list, short col);
BOOL				EmptyList					(struct List *list);



/**/
/*                         - SystemListBrowser -								*/
/**/


/**/
/*                        SystemListBrowserInit()								*/
/**/
void SystemListBrowserInit (void)
{
	/*------------------------------------------*/
	/* Start with disabling the Watcher window. */
	/*------------------------------------------*/
	SleepWindow (Win, &SleepReq);

	/*----------------------------*/
	/* Create nodes in the lists. */
	/*----------------------------*/
	CollectAllLists ();

	/*------------------*/
	/* Open the window. */
	/*------------------*/
	AllocSLBWindow ();
}

/**/
/*                        SystemListBrowserExit()								*/
/**/
void SystemListBrowserExit (void)
{
	FreeAllLists ();
	FreeSLBWindow ();	
	WakeUpWindow (Win, &SleepReq);
}

/**/
/*                           AllocSLBWindow()									*/
/**/
void AllocSLBWindow (void)
{
	/*-------------*/
	/* L O C A L S */
	/*-------------*/
	UBYTE *z = XGetString (M_MENU_SLB_SLB);

	/*--------------*/
	/* Open window. */
	/*--------------*/
	if (!(SLBWin = OpenWindowTags (NULL,
						WA_Left,				Win->LeftEdge,
						WA_Top,				Win->TopEdge + Win->Height,
						WA_Width,			Win->Width,
						WA_Height,			200,
						WA_MinWidth,		160,
						WA_MinHeight,		PubScr->WBorTop + PubScr->RastPort.TxHeight + 1 + 100,
						WA_MaxWidth,		-1,
						WA_MaxHeight,		-1,
						WA_IDCMP,			SLBWIN_IDCMP,
						WA_Flags,			SLBWIN_FLAGS,
						WA_Gadgets,			NULL,
						WA_Title,			z,
						WA_PubScreen,		PubScr,
						WA_NewLookMenus,	TRUE,
						TAG_DONE)))
	{
		CleanExit (RETURN_FAIL, XGetString (M_COULDNT_OPEN_WINDOW), 0);
	}

	/*--------------*/
	/* Create menu. */
	/*--------------*/
	CreateSLBMenu ();

	/*-----------------*/
	/* Create gadgets. */
	/*-----------------*/
	CreateSLBGadgets ();
}

/**/
/*                            FreeSLBWindow()									*/
/**/
void FreeSLBWindow (void)
{
	FreeSLBGadgets ();
	if (SLBMenus)
	{
		FreeSLBMenu ();
	}
	if (SLBWin)
	{
		CloseWindow (SLBWin);
		SLBWin = NULL;
	}
}

/**/
/*                          CreateSLBGadgets()									*/
/**/
void CreateSLBGadgets (void)
{
	/*-------------*/
	/* L O C A L S */
	/*-------------*/
	struct NewGadget	ng;
	struct Gadget		*g;
	UWORD					gc = 0;
	UWORD					GadTopStart;
	UBYTE					*UpdateStr = XGetString (M_UPDATE);

	/*---------*/
	/* C O D E */
	/*---------*/
	/*------------------------*/
	/* Create gadget context. */
	/*------------------------*/
	g = CreateContext (&SLBGList);

	/*-------------------*/
	/* Creating gadgets. */
	/*-------------------*/
	GadTopStart = PubScr->WBorTop + PubScr->RastPort.TxHeight + 1 + GD_SLACK;

	/*------------------------*/
	/* Creating cycle gadget. */
	/*------------------------*/
	ng.ng_LeftEdge		= 10;
	ng.ng_TopEdge		= GadTopStart;
	ng.ng_Width			= SLBWin->Width - (4 * GD_SLACK);
	ng.ng_Height		= 12;
	ng.ng_GadgetID		= GD_CYCLE;
	ng.ng_GadgetText	= NULL;
	ng.ng_TextAttr		= &topaz8;
	ng.ng_Flags			= 0;
	ng.ng_VisualInfo	= VisualInfo;

	SLBGads [gc++] = g = CreateGadget (CYCLE_KIND, g, &ng,
												  GTCY_Labels, CycleText,
												  GTCY_Active, DisplayType,
												  TAG_DONE);

	/*---------------------------*/
	/* Creating listview gadget. */
	/*---------------------------*/
	ng.ng_TopEdge		+= g->Height + GD_SLACK;
	ng.ng_Height		= SLBWin->Height - GadTopStart - 44;
	ng.ng_GadgetID		= GD_LISTVIEW;

	switch (DisplayType)
	{
		case OBJ_SCREEN :
		{
			SLBGads [gc++] = g = CreateGadget (LISTVIEW_KIND, g, &ng,
														  GTLV_Labels,	MyScrList,
														  TAG_DONE);
			break;
		}
		case OBJ_LIBRARY :
		{
			SLBGads [gc++] = g = CreateGadget (LISTVIEW_KIND, g, &ng,
														  GTLV_Labels,	MyLibList,
														  TAG_DONE);
			break;
		}
		case OBJ_DEVICE :
		{
			SLBGads [gc++] = g = CreateGadget (LISTVIEW_KIND, g, &ng,
														  GTLV_Labels,	MyDevList,
														  TAG_DONE);
			break;
		}
		case OBJ_TASK :
		{
			SLBGads [gc++] = g = CreateGadget (LISTVIEW_KIND, g, &ng,
														  GTLV_Labels,	MyTskList,
														  TAG_DONE);
			break;
		}
		case OBJ_PORT :
		{
			SLBGads [gc++] = g = CreateGadget (LISTVIEW_KIND, g, &ng,
														  GTLV_Labels,	MyPrtList,
														  TAG_DONE);
			break;
		}
		case OBJ_RESOURCE :
		{
			SLBGads [gc++] = g = CreateGadget (LISTVIEW_KIND, g, &ng,
														  GTLV_Labels,	MyResList,
														  TAG_DONE);
			break;
		}
		case OBJ_FONT :
		{
			SLBGads [gc++] = g = CreateGadget (LISTVIEW_KIND, g, &ng,
														  GTLV_Labels,	MyFntList,
														  TAG_DONE);
			break;
		}
		case OBJ_VOLUME :
		{
			SLBGads [gc++] = g = CreateGadget (LISTVIEW_KIND, g, &ng,
														  GTLV_Labels,	MyVolList,
														  TAG_DONE);
			break;
		}
		case OBJ_ASSIGN :
		{
			SLBGads [gc++] = g = CreateGadget (LISTVIEW_KIND, g, &ng,
														  GTLV_Labels,	MyAssList,
														  TAG_DONE);
			break;
		}
	}

	/*-----------------------*/
	/* Creating BUTTON: Add. */
	/*-----------------------*/
	ng.ng_TopEdge		+= g->Height + GD_SLACK;
	ng.ng_Height		= 12;
	ng.ng_GadgetText	= UpdateStr;
	ng.ng_GadgetID		= GD_BUTTON;

	SLBGads [gc++] = g = CreateGadget (BUTTON_KIND, g, &ng,
												  GT_Underscore, '_',
												  TAG_DONE);

	/*---------------------------*/
	/* Attach gadgets to window. */
	/*---------------------------*/
	AddGList (SLBWin, SLBGList, -1, -1, NULL);
	SetRast (SLBWin->RPort, 0);
	RefreshGList (SLBGList, Win, NULL, -1);
	GT_RefreshWindow (SLBWin, NULL);
	RefreshWindowFrame (SLBWin);
}

/**/
/*                           FreeSLBGadgets()									*/
/**/
void FreeSLBGadgets (void)
{
	if (SLBGList && SLBWin)
	{
		RemoveGList (SLBWin, SLBGList, -1);
		FreeGadgets (SLBGList);
	}
}

/**/
/*                            CreateSLBMenu()									*/
/**/
void CreateSLBMenu (void)
{
	/*-------------*/
	/* L O C A L S */
	/*-------------*/
	struct NewMenu NewMenu [] =					// the user field of the menu, so
	{														// that we can use it to call AG.
		NM_TITLE,	"Project",				NULL,	0,	NULL,	0L,
			NM_ITEM,	"Update All Lists",	"U",	0,	NULL,	0L,
			NM_ITEM,	"About",					"A",	0,	NULL,	0L,
			NM_ITEM,	NM_BARLABEL,			NULL,	0,	NULL,	0L,
			NM_ITEM,	"Quit",					"Q",	0,	NULL,	0L,

		NM_END,			0,						NULL,	0,	NULL,	0L
	};

	/*---------*/
	/* C O D E */
	/*---------*/
	/*--------------------*/
	/* Localize SLB menu. */
	/*--------------------*/
	NewMenu [0].nm_Label		= XGetString (M_MENU_SLB_PROJECT);
	NewMenu [1].nm_Label		= XGetString (M_MENU_SLB_UPDATE_ALL_LISTS);
	NewMenu [1].nm_CommKey	= XGetString (M_MENU_SLB_UPDATE_ALL_LISTS_SC);
	NewMenu [2].nm_Label		= XGetString (M_MENU_SLB_ABOUT);
	NewMenu [2].nm_CommKey	= XGetString (M_MENU_SLB_ABOUT_SC);
	NewMenu [4].nm_Label		= XGetString (M_MENU_SLB_QUIT);
	NewMenu [4].nm_CommKey	= XGetString (M_MENU_SLB_QUIT_SC);

	if (!(SLBMenus = CreateMenus (NewMenu,
											GTMN_NewLookMenus,	TRUE,
											GTMN_FullMenu,			TRUE,
											TAG_DONE)))
	{
		CleanExit (RETURN_FAIL, XGetString (M_UNABLE_TO_CREATE_MENU), 0);
	}
	LayoutMenus (SLBMenus, VisualInfo,
					 GTMN_NewLookMenus,	TRUE,
					 TAG_DONE);
	SetMenuStrip (SLBWin, SLBMenus);
}

/**/
/*                             FreeSLBMenu()										*/
/**/
void FreeSLBMenu (void)
{
	if (SLBMenus)
	{
		ClearMenuStrip (SLBWin);
		FreeMenus (SLBMenus);
	}
}

/**/
/*                            AllocAllLists()									*/
/**/
BOOL AllocAllLists (void)
{
	if (mem_pool_ok)
	{
		if (!(MyScrList = AllocPooled (mem_pool, sizeof (struct List))))
		{
			return (FALSE);
		}
		if (!(MyLibList = AllocPooled (mem_pool, sizeof (struct List))))
		{
			return (FALSE);
		}
		if (!(MyDevList = AllocPooled (mem_pool, sizeof (struct List))))
		{
			return (FALSE);
		}
		if (!(MyTskList = AllocPooled (mem_pool, sizeof (struct List))))
		{
			return (FALSE);
		}
		if (!(MyPrtList = AllocPooled (mem_pool, sizeof (struct List))))
		{
			return (FALSE);
		}
		if (!(MyResList = AllocPooled (mem_pool, sizeof (struct List))))
		{
			return (FALSE);
		}
		if (!(MyFntList = AllocPooled (mem_pool, sizeof (struct List))))
		{
			return (FALSE);
		}
		if (!(MyVolList = AllocPooled (mem_pool, sizeof (struct List))))
		{
			return (FALSE);
		}
		if (!(MyAssList = AllocPooled (mem_pool, sizeof (struct List))))
		{
			return (FALSE);
		}
	}
	else
	{
		if (!(MyScrList = AllocVec (sizeof (struct List), MEMF_ANY | MEMF_CLEAR)))
		{
			return (FALSE);
		}
		if (!(MyLibList = AllocVec (sizeof (struct List), MEMF_ANY | MEMF_CLEAR)))
		{
			return (FALSE);
		}
		if (!(MyDevList = AllocVec (sizeof (struct List), MEMF_ANY | MEMF_CLEAR)))
		{
			return (FALSE);
		}
		if (!(MyTskList = AllocVec (sizeof (struct List), MEMF_ANY | MEMF_CLEAR)))
		{
			return (FALSE);
		}
		if (!(MyPrtList = AllocVec (sizeof (struct List), MEMF_ANY | MEMF_CLEAR)))
		{
			return (FALSE);
		}
		if (!(MyResList = AllocVec (sizeof (struct List), MEMF_ANY | MEMF_CLEAR)))
		{
			return (FALSE);
		}
		if (!(MyFntList = AllocVec (sizeof (struct List), MEMF_ANY | MEMF_CLEAR)))
		{
			return (FALSE);
		}
		if (!(MyVolList = AllocVec (sizeof (struct List), MEMF_ANY | MEMF_CLEAR)))
		{
			return (FALSE);
		}
		if (!(MyAssList = AllocVec (sizeof (struct List), MEMF_ANY | MEMF_CLEAR)))
		{
			return (FALSE);
		}
	}
}

/**/
/*                            ResetAllLists()									*/
/**/
void ResetAllLists (void)
{
	NewList (MyScrList);
	NewList (MyLibList);
	NewList (MyDevList);
	NewList (MyTskList);
	NewList (MyPrtList);
	NewList (MyResList);
	NewList (MyFntList);
	NewList (MyVolList);
	NewList (MyAssList);
}

/**/
/*                             RemAllNodes()										*/
/**/
void RemAllNodes (void)
{
	if (MyScrList)
	{
		FreeAllNodes (MyScrList);
	}
	if (MyLibList)
	{
		FreeAllNodes (MyLibList);
	}
	if (MyDevList)
	{
		FreeAllNodes (MyDevList);
	}
	if (MyTskList)
	{
		FreeAllNodes (MyTskList);
	}
	if (MyPrtList)
	{
		FreeAllNodes (MyPrtList);
	}
	if (MyResList)
	{
		FreeAllNodes (MyResList);
	}
	if (MyFntList)
	{
		FreeAllNodes (MyFntList);
	}
	if (MyVolList)
	{
		FreeAllNodes (MyVolList);
	}
	if (MyAssList)
	{
		FreeAllNodes (MyAssList);
	}
}

/**/
/*                            FreeAllLists()										*/
/**/
void FreeAllLists (void)
{
	if (!mem_pool_ok)
	{
		if (MyScrList)
		{
			FreeVec (MyScrList);
		}
		if (MyLibList)
		{
			FreeVec (MyLibList);
		}
		if (MyDevList)
		{
			FreeVec (MyDevList);
		}
		if (MyTskList)
		{
			FreeVec (MyTskList);
		}
		if (MyPrtList)
		{
			FreeVec (MyPrtList);
		}
		if (MyResList)
		{
			FreeVec (MyResList);
		}
		if (MyFntList)
		{
			FreeVec (MyFntList);
		}
		if (MyVolList)
		{
			FreeVec (MyVolList);
		}
		if (MyAssList)
		{
			FreeVec (MyAssList);
		}
	}
}

/**/
/*                           CollectAllLists()									*/
/**/
void CollectAllLists (void)
{
	if (SLBWin)
	{
		SleepWindow (SLBWin, &SleepReq2);
	}
	if (EmptyList (MyScrList))
	{
		CollectScrList ();
	}
	if (EmptyList (MyLibList))
	{
		CollectLibList ();
	}
	if (EmptyList (MyDevList))
	{
		CollectDevList ();
	}
	if (EmptyList (MyTskList))
	{
		CollectTskList ();
	}
	if (EmptyList (MyPrtList))
	{
		CollectPrtList ();
	}
	if (EmptyList (MyResList))
	{
		CollectResList ();
	}
	if (SLBWin)
	{
		WakeUpWindow (SLBWin, &SleepReq2);
	}
}

/**/
/*                           CollectScrList()									*/
/**/
void CollectScrList (void)
{
	/*-------------*/
	/* L O C A L S */
	/*-------------*/
	struct List				*sl;
	struct PubScreenNode	*psn;
	struct MyScrNode		*msn;

	/*---------*/
	/* C O D E */
	/*---------*/
	sl = LockPubScreenList ();
	for (psn = (struct PubScreenNode *) sl->lh_Head; psn->psn_Node.ln_Succ; psn = (struct PubScreenNode *) psn->psn_Node.ln_Succ)
	{
		if (mem_pool_ok)
		{
			msn = AllocPooled (mem_pool, sizeof (struct MyScrNode));
		}
		else
		{
			msn = AllocVec (sizeof (struct MyScrNode), MEMF_ANY | MEMF_CLEAR);
		}
		if (msn)
		{
			strncpy (msn->msn_PubName, psn->psn_Node.ln_Name, MAXPUBSCREENNAME);
			strncpy (msn->msn_TitName, psn->psn_Screen->Title, NSIZE);
			msn->msn_Node.ln_Name= msn->msn_PubName;
			msn->msn_Address		= psn->psn_Screen;
			msn->msn_Box.Left		= psn->psn_Screen->LeftEdge;
			msn->msn_Box.Top		= psn->psn_Screen->TopEdge;
			msn->msn_Box.Width	= psn->psn_Screen->Width;
			msn->msn_Box.Height	= psn->psn_Screen->Height;
			msn->msn_Depth			= psn->psn_Screen->RastPort.BitMap->Depth;
			msn->msn_Font			= psn->psn_Screen->Font;
			msn->msn_Windows		= (psn->psn_Screen->FirstWindow) ?
										  TRUE : FALSE;
			msn->msn_Gadgets		= (psn->psn_Screen->FirstGadget) ?
										  TRUE : FALSE;
			CreateANode (MyScrList, (struct Node *) msn);
		}
	}
	UnlockPubScreenList ();

	SortList (MyScrList, 0);
}

/**/
/*                           CollectLibList()									*/
/**/
void CollectLibList (void)
{
	/*-------------*/
	/* L O C A L S */
	/*-------------*/
	struct List				*liblist;
	struct Library			*n;
	struct MyLibDevNode	*mldn;

	/*---------*/
	/* C O D E */
	/*---------*/
	liblist = &(SysBase->LibList);
	for (n = (struct Library *) liblist->lh_Head; n->lib_Node.ln_Succ; n = (struct Library *) n->lib_Node.ln_Succ)
	{
		if (mem_pool_ok)
		{
			mldn = AllocPooled (mem_pool, sizeof (struct MyLibDevNode));
		}
		else
		{
			mldn = AllocVec (sizeof (struct MyLibDevNode), MEMF_ANY | MEMF_CLEAR);
		}
		if (mldn)
		{
			if (n->lib_Node.ln_Name [0] != '\0')
			{
				strncpy (mldn->mldn_Name, n->lib_Node.ln_Name, NSIZE);
			}
			else
			{
				strncpy (mldn->mldn_Name, "??", NSIZE);
			}
			mldn->mldn_Node.ln_Name	= mldn->mldn_Name;
			mldn->mldn_Version		= n->lib_Version;
			mldn->mldn_Revision		= n->lib_Revision;
			if (n->lib_IdString)
			{
				strncpy (mldn->mldn_IdString, n->lib_IdString, NSIZE);
			}
			else
			{
				strncpy (mldn->mldn_IdString, "??", NSIZE);
			}
			mldn->mldn_OpenCount		= n->lib_OpenCnt;
			CreateANode (MyLibList, (struct Node *) mldn);
		}
	}
	SortList (MyLibList, 0);
}

/**/
/*                           CollectDevList()									*/
/**/
void CollectDevList (void)
{
	/*-------------*/
	/* L O C A L S */
	/*-------------*/
	struct List				*devlist;
	struct Library			*n;
	struct MyLibDevNode	*mldn;

	/*---------*/
	/* C O D E */
	/*---------*/
	Forbid ();
	devlist = &(SysBase->DeviceList);
	for (n = (struct Library *) devlist->lh_Head; n->lib_Node.ln_Succ; n = (struct Library *) n->lib_Node.ln_Succ)
	{
		if (mem_pool_ok)
		{
			mldn = AllocPooled (mem_pool, sizeof (struct MyLibDevNode));
		}
		else
		{
			mldn = AllocVec (sizeof (struct MyLibDevNode), MEMF_ANY | MEMF_CLEAR);
		}
		if (mldn)
		{
			if (n->lib_Node.ln_Name [0] != '\0')
			{
				strncpy (mldn->mldn_Name, n->lib_Node.ln_Name, NSIZE);
			}
			else
			{
				strncpy (mldn->mldn_Name, "??", NSIZE);
			}
			mldn->mldn_Node.ln_Name	= mldn->mldn_Name;
			mldn->mldn_Version		= n->lib_Version;
			mldn->mldn_Revision		= n->lib_Revision;
			if (n->lib_IdString)
			{
				strncpy (mldn->mldn_IdString, n->lib_IdString, NSIZE);
			}
			else
			{
				strncpy (mldn->mldn_IdString, "??", NSIZE);
			}
			mldn->mldn_OpenCount		= n->lib_OpenCnt;
			CreateANode (MyDevList, (struct Node *) mldn);
		}
	}
	Permit ();
	SortList (MyDevList, 0);
}

/**/
/*                           CollectTskList()									*/
/**/
void CollectTskList (void)
{
	/*-------------*/
	/* L O C A L S */
	/*-------------*/
	struct List							*tsklist;
	struct Task							*n;
	struct MyTskNode					*mtn;
	struct CommandLineInterface	*cli;
	UBYTE									*ptrName;
	LONG									len;

	/*---------*/
	/* C O D E */
	/*---------*/
	/*-----------------------------------------------------------*/
	/* Disable task switching while we're peeking the task list. */
	/*-----------------------------------------------------------*/
	Forbid ();

	/*-----------------------------------------------*/
	/* Run through the task list maintained by Exec. */
	/*-----------------------------------------------*/
	tsklist = &(SysBase->TaskWait);
	for (n = (struct Task *) tsklist->lh_Head; n->tc_Node.ln_Succ; n = (struct Task *) n->tc_Node.ln_Succ)
	{
		/*-----------------------------------*/
		/* Check if we can use the V39 pool. */
		/*-----------------------------------*/
		if (mem_pool_ok)
		{
			mtn = AllocPooled (mem_pool, sizeof (struct MyTskNode));
		}
		else
		{
			mtn = AllocVec (sizeof (struct MyTskNode), MEMF_ANY | MEMF_CLEAR);
		}

		/*-------------------------------------------------*/
		/* Check if we could allocate memory for our node. */
		/*-------------------------------------------------*/
		if (mtn)
		{
			/*------------------------------------*/
			/* Check that the task name is valid. */
			/*------------------------------------*/
			if ((n->tc_Node.ln_Name [0] != '\0') && (n->tc_Node.ln_Name [0] != ' '))
			{
				strncpy (mtn->mtn_Name, n->tc_Node.ln_Name, NSIZE);
				mtn->mtn_Priority = n->tc_Node.ln_Pri;

				/*--------------------------------------*/
				/* Check if this task is a CLI process. */
				/*--------------------------------------*/
				if ((((struct Process *) n)->pr_TaskNum) && (n->tc_Node.ln_Type == NT_PROCESS))
				{
					/*---------------------------------*/
					/* Remember this is a CLI process. */
					/*---------------------------------*/
					mtn->mtn_CliProcess	= TRUE;

					/*-----------------------------*/
					/* Get the CLI process number. */
					/*-----------------------------*/
					mtn->mtn_CliNumber	= ((struct Process *) n)->pr_TaskNum;

					/*-------------------------------------------*/
					/* Check that we have a valid CLI structure. */
					/*-------------------------------------------*/
					if (((struct Process *) n)->pr_CLI)
					{
						if (cli = (struct CommandLineInterface *) BADDR (((struct Process *) n)->pr_CLI))
						{
							/*------------------------------*/
							/* Get the 'real' command name. */
							/*------------------------------*/
							ptrName = (BYTE *) BADDR (cli->cli_CommandName);
							if (isalnum (ptrName [1]))
							{
								/*-------------------------------------------------------------------------*/
								/* Some processes really fake themselves as a CLI process and don't really */
								/* have a valid CommandName, this we have to take care of (sigh!).			*/
								/*-------------------------------------------------------------------------*/
								len = *ptrName++;
								strncpy (mtn->mtn_CliCmdName, ptrName, len);
								mtn->mtn_CliCmdName [len] = '\0';

								CopyMem (cli, &mtn->mtn_Cli, sizeof (struct CommandLineInterface));

								mtn->mtn_Stack	= (void *) BADDR (((struct Process *) n)->pr_StackBase);
								mtn->mtn_Lower	= n->tc_SPLower;
								mtn->mtn_Upper	= n->tc_SPReg;
							}
							else
							{
								strcpy (mtn->mtn_CliCmdName, mtn->mtn_Name);

								/*-------------------------*/
								/* Get the stack settings. */
								/*-------------------------*/
								mtn->mtn_Stack	= n->tc_SPReg;
								mtn->mtn_Lower	= n->tc_SPLower;
								mtn->mtn_Upper	= n->tc_SPUpper;
							}
						}
						else
						{
							strcpy (mtn->mtn_CliCmdName, mtn->mtn_Name);

							/*-------------------------*/
							/* Get the stack settings. */
							/*-------------------------*/
							mtn->mtn_Stack	= n->tc_SPReg;
							mtn->mtn_Lower	= n->tc_SPLower;
							mtn->mtn_Upper	= n->tc_SPUpper;
						}
					}
					else
					{
						strcpy (mtn->mtn_CliCmdName, mtn->mtn_Name);

						/*-------------------------*/
						/* Get the stack settings. */
						/*-------------------------*/
						mtn->mtn_Stack	= n->tc_SPReg;
						mtn->mtn_Lower	= n->tc_SPLower;
						mtn->mtn_Upper	= n->tc_SPUpper;
					}
				}
				else
				{
					mtn->mtn_CliProcess = FALSE;

					/*-------------------------*/
					/* Get the stack settings. */
					/*-------------------------*/
					mtn->mtn_Stack	= n->tc_SPReg;
					mtn->mtn_Lower	= n->tc_SPLower;
					mtn->mtn_Upper	= n->tc_SPUpper;
				}
			}
			else
			{
				mtn->mtn_CliProcess = FALSE;
				strncpy (mtn->mtn_Name, ">Unknown<", NSIZE);

				/*-------------------------*/
				/* Get the stack settings. */
				/*-------------------------*/
				mtn->mtn_Stack	= n->tc_SPReg;
				mtn->mtn_Lower	= n->tc_SPLower;
				mtn->mtn_Upper	= n->tc_SPUpper;
			}

			/*------------------------------------------------------------------------------*/
			/* If task is a CLI process with a 'real' name, we will choose this name as our */
			/* node name (so that it goes with this name in the listview!).					  */
			/*------------------------------------------------------------------------------*/
			if (mtn->mtn_CliProcess && mtn->mtn_CliCmdName [0] != '\0' &&
				 mtn->mtn_CliCmdName [0] != ' ')
			{
				mtn->mtn_Node.ln_Name = mtn->mtn_CliCmdName;
			}
			else
			{
				mtn->mtn_Node.ln_Name = mtn->mtn_Name;
			}

			/*-----------------------*/
			/* Create the task node. */
			/*-----------------------*/
			CreateANode (MyTskList, (struct Node *) mtn);
		}
	}

	/*----------------------------------------------------------------------*/
	/* We're done peeking the task list, so we enable task switching again. */
	/*----------------------------------------------------------------------*/
	Permit ();

	/*----------------------------------------------*/
	/* End with sorting the nodes in our task list. */
	/*----------------------------------------------*/
	SortList (MyTskList, 0);
}

/**/
/*                           CollectPrtList()									*/
/**/
void CollectPrtList (void)
{
	/*-------------*/
	/* L O C A L S */
	/*-------------*/
	struct List			*portlist;
	struct Node			*n;
	struct MyPrtNode	*mpn;

	/*---------*/
	/* C O D E */
	/*---------*/
	Forbid ();
	portlist = &(SysBase->PortList);
	for (n = portlist->lh_Head; n->ln_Succ; n = n->ln_Succ)
	{
		if (mem_pool_ok)
		{
			mpn = AllocPooled (mem_pool, sizeof (struct MyPrtNode));
		}
		else
		{
			mpn = AllocVec (sizeof (struct MyPrtNode), MEMF_ANY | MEMF_CLEAR);
		}
		if (mpn)
		{
			if (n->ln_Name != '\0')
			{
				strncpy (mpn->mpn_Name, n->ln_Name, NSIZE);
			}
			else
			{
				strncpy (mpn->mpn_Name, "??", NSIZE);
			}
			mpn->mpn_Node.ln_Name	= mpn->mpn_Name;
			mpn->mpn_Pri				= n->ln_Pri;
			mpn->mpn_SigTask			= ((struct MsgPort *) n)->mp_SigTask;
			mpn->mpn_Address			= n;
			CreateANode (MyPrtList, (struct Node *) mpn);
		}
	}
	Permit ();
	SortList (MyPrtList, 0);
}

/**/
/*                           CollectResList()									*/
/**/
void CollectResList (void)
{
	/*-------------*/
	/* L O C A L S */
	/*-------------*/
	struct List			*reslist;
	struct Node			*n;
	struct MyResNode	*mrn;

	/*---------*/
	/* C O D E */
	/*---------*/
	Forbid ();
	reslist = &(SysBase->ResourceList);
	for (n = reslist->lh_Head; n->ln_Succ; n = n->ln_Succ)
	{
		if (mem_pool_ok)
		{
			mrn = AllocPooled (mem_pool, sizeof (struct MyResNode));
		}
		else
		{
			mrn = AllocVec (sizeof (struct MyResNode), MEMF_ANY | MEMF_CLEAR);
		}
		if (mrn)
		{
			if (n->ln_Name != '\0')
			{
				strncpy (mrn->mrn_Name, n->ln_Name, NSIZE);
			}
			else
			{
				strncpy (mrn->mrn_Name, "??", NSIZE);
			}
			mrn->mrn_Node.ln_Name	= mrn->mrn_Name;
			if (mrn->mrn_Version = ((struct Library *) n)->lib_Version)
			{
				mrn->mrn_Revision			= ((struct Library *) n)->lib_Revision;
				if (((struct Library *) n)->lib_IdString)
				{
					strncpy (mrn->mrn_IdString, ((struct Library *) n)->lib_IdString, NSIZE);
				}
				else
				{
					strncpy (mrn->mrn_IdString, "??", NSIZE);
				}
				mrn->mrn_OpenCount		= ((struct Library *) n)->lib_OpenCnt;
			}
			CreateANode (MyResList, (struct Node *) mrn);
		}
	}
	Permit ();
	SortList (MyResList, 0);
}

/**/
/*                           CollectFntList()									*/
/**/
void CollectFntList (void)
{
	/*-------------*/
	/* L O C A L S */
	/*-------------*/
	struct MyFntNode			*mfn;
	LONG							size, shortage;
	LONG							x;
	struct AvailFontsHeader	*afh = NULL;
	struct TAvailFonts		*afont;
	struct TextFont			*myfont;
	ULONG							dpi;
	UBYTE							tmpbuf [24];
	BOOL							printed;

	/*---------*/
	/* C O D E */
	/*---------*/
	/*---------------------------------------------*/
	/* Get the available fonts with a buffer of 0. */
	/*---------------------------------------------*/
	size = AvailFonts ((STRPTR) afh, 0L, AFF_MEMORY | AFF_DISK | AFF_SCALED | AFF_TAGGED);

	/*--------------------------------------------------------*/
	/* Loop until we have enough buffer for AvailFontsHeader. */
	/*--------------------------------------------------------*/
	do
	{
		/*---------------------------------------*/
		/* Allocate memory for AvailFontsHeader. */
		/*---------------------------------------*/
		if (mem_pool_ok)
		{
			afh = AllocPooled (mem_pool, size);
		}
		else
		{
			afh = AllocVec (size, MEMF_ANY);
		}
		if (afh)
		{
			/*-------------------------------------------------------------------------*/
			/* Get the available fonts with the purpose of getting the number of bytes */
			/* to allocate.																				*/
			/*-------------------------------------------------------------------------*/
			shortage = AvailFonts ((STRPTR) afh, size,
										  AFF_MEMORY | AFF_DISK | AFF_SCALED | AFF_TAGGED);

			/*----------------------*/
			/* Is there a shortage? */
			/*----------------------*/
			if (shortage)
			{
				/*-------------------------------------------------------------------------------*/
				/* Free the previously allocated memory, and add the shortage to our buffersize. */
				/*-------------------------------------------------------------------------------*/
				if (!mem_pool_ok)
				{
					FreeVec (afh);
				}
				afh = NULL;
				size += shortage;
			}
		}
		else
		{
			CleanExit (RETURN_FAIL, NULL, ERROR_NO_FREE_STORE);
		}
	} while (shortage > 0);

	/*-------------------------------------------------*/
	/* Check that we actually have a AvailFontsHeader. */
	/*-------------------------------------------------*/
	if (afh)
	{
		/*---------------------*/
		/* Get the first font. */
		/*---------------------*/
		afont = (struct TAvailFonts *) &(afh [1]);

		/*------------------------*/
		/* Run through the fonts. */
		/*------------------------*/
		for (x = 0; x < afh->afh_NumEntries; x++)
		{
			if (myfont = OpenDiskFont ((struct TextAttr *) &(afont->taf_Attr)))
			{
				/*-------------------------------*/
				/* Allocate memory for the node. */
				/*-------------------------------*/
				if (mem_pool_ok)
				{
					mfn = AllocPooled (mem_pool, sizeof (struct MyFntNode));
				}
				else
				{
					mfn = AllocVec (sizeof (struct MyFntNode), MEMF_ANY | MEMF_CLEAR);
				}
				if (!(mfn))
				{
					CloseFont (myfont);
					if (!mem_pool_ok)
					{
						FreeVec (afh);
					}
					CleanExit (RETURN_FAIL, NULL, ERROR_NO_FREE_STORE);
				}

				/*----------------------*/
				/* Get the font's name. */
				/*----------------------*/
				sprintf (mfn->mfn_Name, "%-28s %3ld\0",
												(UBYTE *) myfont->tf_Message.mn_Node.ln_Name,
												myfont->tf_YSize);
				mfn->mfn_Node.ln_Name = mfn->mfn_Name;

				/*--------------------------------------------*/
				/* Copy the TextFont structure into our node. */
				/*--------------------------------------------*/
				CopyMem (myfont, &mfn->mfn_tf, sizeof (struct TextFont));

				/*-----------------------------------------------*/
				/* Try to get extra information about X & Y dpi. */
				/*-----------------------------------------------*/
				dpi = GetTagData (TA_DeviceDPI, 0L, ((struct TextFontExtension *)
																(myfont->tf_Extension))->tfe_Tags);
				if (dpi)
				{
					/*----------------------------------------------------------------*/
					/* DPI tags are available for this font, so we put the DPI values */
					/* into our node.																	*/
					/*----------------------------------------------------------------*/
					sprintf (tmpbuf, "%d\0", ((dpi & 0xFFFF0000) >> 16));
					mfn->mfn_XDPI = atoi (tmpbuf);
					sprintf (tmpbuf, "%d\0", (dpi & 0x0000FFFF));
					mfn->mfn_YDPI = atoi (tmpbuf);
				}
				else
				{
					/*--------------------*/
					/* No DPI tags found. */
					/*--------------------*/
					mfn->mfn_XDPI = 0;
					mfn->mfn_YDPI = 0;
				}

				/*---------------------*/
				/* Check type of font. */
				/*---------------------*/
				switch (afont->taf_Type)
				{
					case AFF_MEMORY :
					{
						strcpy (mfn->mfn_Type, "Memory");
						break;
					}
					case AFF_DISK :
					{
						strcpy (mfn->mfn_Type, "Disk");
						break;
					}
					case AFF_SCALED :
					{
						strcpy (mfn->mfn_Type, "Scaled");
						break;
					}
					case AFF_BITMAP :
					{
						strcpy (mfn->mfn_Type, "Bitmap");
						break;
					}
					case AFF_TAGGED :
					{
						strcpy (mfn->mfn_Type, "Tagged");
						break;
					}
				}

				printed = FALSE;
				mfn->mfn_StylesText [0] = '\0';

				/*----------------------------------------------*/
				/* Build string that corresponds to font style. */
				/*----------------------------------------------*/
				if ((myfont->tf_Style & FSF_ITALIC) == FSF_ITALIC)
				{
					if (printed)
					{
						strcat (mfn->mfn_StylesText, ", ");
					}
					strcat (mfn->mfn_StylesText, "Italic");
					printed = TRUE;
				}
				if ((myfont->tf_Style & FSF_BOLD) == FSF_BOLD)
				{
					if (printed)
					{
						strcat (mfn->mfn_StylesText, ", ");
					}
					strcat (mfn->mfn_StylesText, "Bold");
					printed = TRUE;
				}
				if ((myfont->tf_Style & FSF_UNDERLINED) == FSF_UNDERLINED)
				{
					if (printed)
					{
						strcat (mfn->mfn_StylesText, ", ");
					}
					strcat (mfn->mfn_StylesText, "Underlined");
					printed = TRUE;
				}
				if ((myfont->tf_Style & FSF_EXTENDED) == FSF_EXTENDED)
				{
					if (printed)
					{
						strcat (mfn->mfn_StylesText, ", ");
					}
					strcat (mfn->mfn_StylesText, "Extended");
					printed = TRUE;
				}
				if ((myfont->tf_Style & FSF_COLORFONT) == FSF_COLORFONT)
				{
					if (printed)
					{
						strcat (mfn->mfn_StylesText, ", ");
					}
					strcat (mfn->mfn_StylesText, "Color");
					printed = TRUE;
				}
				if ((myfont->tf_Style & FSF_TAGGED) == FSF_TAGGED)
				{
					if (printed)
					{
						strcat (mfn->mfn_StylesText, ", ");
					}
					strcat (mfn->mfn_StylesText, "Tagged");
					printed = TRUE;
				}
				if (!printed)
				{
					strcat (mfn->mfn_StylesText, "No style bits set.");
				}
				else
				{
					strcat (mfn->mfn_StylesText, ".");
				}

				printed = FALSE;
				mfn->mfn_FlagsText [0] = '\0';

				/*----------------------------------------------*/
				/* Build string that corresponds to font flags. */
				/*----------------------------------------------*/
				if ((myfont->tf_Flags & FPF_ROMFONT) == FPF_ROMFONT)
				{
					strcat (mfn->mfn_FlagsText, "RomFont");
					printed = TRUE;
				}
				if ((myfont->tf_Flags & FPF_DISKFONT) == FPF_DISKFONT)
				{
					if (printed)
					{
						strcat (mfn->mfn_FlagsText, ", ");
					}
					strcat (mfn->mfn_FlagsText, "DiskFont");
					printed = TRUE;
				}
				if ((myfont->tf_Flags & FPF_REVPATH) == FPF_REVPATH)
				{
					if (printed)
					{
						strcat (mfn->mfn_FlagsText, ", ");
					}
					strcat (mfn->mfn_FlagsText, "Reversed Path");
					printed = TRUE;
				}
				if ((myfont->tf_Flags & FPF_TALLDOT) == FPF_TALLDOT)
				{
					if (printed)
					{
						strcat (mfn->mfn_FlagsText, ", ");
					}
					strcat (mfn->mfn_FlagsText, "Tall Dots");
					printed = TRUE;
				}
				if ((myfont->tf_Flags & FPF_WIDEDOT) == FPF_WIDEDOT)
				{
					if (printed)
					{
						strcat (mfn->mfn_FlagsText, ", ");
					}
					strcat (mfn->mfn_FlagsText, "Wide Dots");
					printed = TRUE;
				}
				if ((myfont->tf_Flags & FPF_PROPORTIONAL) == FPF_PROPORTIONAL)
				{
					if (printed)
					{
						strcat (mfn->mfn_FlagsText, ", ");
					}
					strcat (mfn->mfn_FlagsText, "Proportional");
					printed = TRUE;
				}
				if ((myfont->tf_Flags & FPF_DESIGNED) == FPF_DESIGNED)
				{
					if (printed)
					{
						strcat (mfn->mfn_FlagsText, ", ");
					}
					strcat (mfn->mfn_FlagsText, "Designed");
					printed = TRUE;
				}
				if ((myfont->tf_Flags & FPF_REMOVED) == FPF_REMOVED)
				{
					if (printed)
					{
						strcat (mfn->mfn_FlagsText, ", ");
					}
					strcat (mfn->mfn_FlagsText, "Removed");
					printed = TRUE;
				}
				if (!printed)
				{
					strcat (mfn->mfn_FlagsText, "No flag bits set.");
				}
				else
				{
					strcat (mfn->mfn_FlagsText, ".");
				}

				/*-----------------------------------------------------------------*/
				/* Here we finally create the node and close the font we examined. */
				/*-----------------------------------------------------------------*/
				CreateANode (MyFntList, (struct Node *) mfn);
				CloseFont (myfont);
			}

			/*-----------------------*/
			/* Advance to next font. */
			/*-----------------------*/
			afont++;
		}
		/*------------------------*/
		/* Free AvailFontsHeader. */
		/*------------------------*/
		if (!mem_pool_ok)
		{
			FreeVec (afh);
		}
	}
	SortList (MyFntList, 0);
}

/**/
/*                           CollectVolList()									*/
/**/
void CollectVolList (void)
{
	/*-------------*/
	/* L O C A L S */
	/*-------------*/
	struct MyVolNode				*mvn;
	struct MyAssNode				*man;
	struct RootNode				*RootNode;
	struct DosInfo					*DosInfo;
	struct DeviceList				*DeviceList;
	UBYTE								*ptrName;
	LONG								len;
	BPTR								VLock;
	struct InfoData __aligned	vid;
	struct ClockData				ClockData;
	BOOL								GoOn;

	/*---------*/
	/* C O D E */
	/*---------*/
	/*-------------------------------------------*/
	/* Initial steps for walking the DeviceList. */
	/*-------------------------------------------*/
	RootNode		= (struct RootNode *)	DOSBase->dl_Root;
	DosInfo		= (struct DosInfo *)		BADDR (RootNode->rn_Info);
	DeviceList	= (struct DeviceList *) BADDR (DosInfo->di_DevInfo);

	/*------------------------*/
	/* Walking the DeviceList */
	/*------------------------*/
	while (DeviceList)
	{
		ptrName = (BYTE *) BADDR (DeviceList->dl_Name);
		len = *ptrName++;

		/*---------------------------------------------------------------*/
		/* Check that device is a volume and that everything else is ok. */
		/*---------------------------------------------------------------*/
		if ((len <= NSIZE)								&&
			 (DeviceList->dl_Type == DLT_VOLUME)	&&
			 (DeviceList->dl_Task))
		{
			/*-------------------------------*/
			/* Allocate memory for our node. */
			/*-------------------------------*/
			if (mem_pool_ok)
			{
				mvn = AllocPooled (mem_pool, sizeof (struct MyVolNode));
			}
			else
			{
				mvn = AllocVec (sizeof (struct MyVolNode), MEMF_ANY | MEMF_CLEAR);
			}
			if (!mvn)
			{
				CleanExit (RETURN_FAIL, NULL, ERROR_NO_FREE_STORE);
			}

			/*-------------*/
			/* Build node. */
			/*-------------*/
			/*------------------*/
			/* Get volume name. */
			/*------------------*/
			strcpy (mvn->mvn_Name, ptrName);
			strcat (mvn->mvn_Name, ":");
			mvn->mvn_Node.ln_Name = mvn->mvn_Name;

			/*---------------------------*/
			/* Get volume creation date. */
			/*---------------------------*/
			Amiga2Date ((DeviceList->dl_VolumeDate.ds_Days * 24 * 60 * 60) +
							(DeviceList->dl_VolumeDate.ds_Minute * 60),
							&ClockData);
			sprintf (mvn->mvn_CrDate, "%02d-%02d-%4d %02d:%02d\0",
						ClockData.mday, ClockData.month, ClockData.year,
						ClockData.hour, ClockData.min);

			if (!(VLock = Lock (mvn->mvn_Name, ACCESS_READ)))
			{
				mvn->mvn_Size = 0;
				mvn->mvn_Free = 0;
			}
			else
			{
				if (!(Info (VLock, &vid)))
				{
					mvn->mvn_Size = 0;
					mvn->mvn_Free = 0;
				}
				else
				{
					mvn->mvn_Size = vid.id_NumBlocks * vid.id_BytesPerBlock;
					mvn->mvn_Free = (vid.id_NumBlocks - vid.id_NumBlocksUsed) *
										  vid.id_BytesPerBlock;
				}
				if (!(NameFromLock (VLock, mvn->mvn_DevName, NSIZE)))
				{
					mvn->mvn_DevName [0] = '\0';
				}
				UnLock (VLock);
			}

			/*--------------------------*/
			/* And the create the node. */
			/*--------------------------*/
			CreateANode (MyVolList, (struct Node *) mvn);
		}

		/*----------------------------------------------------------------*/
		/* Check that device is an assign and that everything else is ok. */
		/*----------------------------------------------------------------*/
		if (len <= NSIZE																								&&
			 ((DeviceList->dl_Type == DLT_DIRECTORY) || (DeviceList->dl_Type == DLT_LATE) ||
			  (DeviceList->dl_Type == DLT_NONBINDING))														&&
			 DeviceList->dl_Task)
		{
			/*-------------------------------*/
			/* Allocate memory for our node. */
			/*-------------------------------*/
			if (mem_pool_ok)
			{
				man = AllocPooled (mem_pool, sizeof (struct MyAssNode));
			}
			else
			{
				man = AllocVec (sizeof (struct MyAssNode), MEMF_ANY | MEMF_CLEAR);
			}
			if (!man)
			{
				CleanExit (RETURN_FAIL, NULL, ERROR_NO_FREE_STORE);
			}

			/*-------------*/
			/* Build node. */
			/*-------------*/
			strcpy (man->man_Name, ptrName);
			strcat (man->man_Name, ":");
			man->man_Node.ln_Name = man->man_Name;

			if (!(VLock = Lock (man->man_Name, ACCESS_READ)))
			{
				man->man_Name [0] = '\0';
			}
			else
			{
				if (!(NameFromLock (VLock, man->man_DirName, NSIZE)))
				{
					man->man_Name [0] = '\0';
				}
				UnLock (VLock);
			}

			/*----------------------------------------------*/
			/* Check if assign is of the late-binding type. */
			/*----------------------------------------------*/
			if (DeviceList->dl_Type != DLT_DIRECTORY)
			{
				man->man_Late = TRUE;
			}
			else
			{
				man->man_Late = FALSE;
			}

			/*--------------------------*/
			/* And the create the node. */
			/*--------------------------*/
			CreateANode (MyAssList, (struct Node *) man);
		}
		DeviceList = (struct DeviceList *) BADDR (DeviceList->dl_Next);
	}

	/*----------------------------------------------------------------------------*/
	/* So far so good. Now, in order to find a volumes device name (say that your */
	/* volume's name is 'Workbench' then the volumes device name could be HD0:),  */
	/* we need to run through the DLT_DEVICE list. But before we do that we need	*/
	/* to disable DOS requesters for PC0: etc.												*/
	/*----------------------------------------------------------------------------*/
	for (mvn = (struct MyVolNode *) MyVolList->lh_Head;
		  mvn->mvn_Node.ln_Succ;
		  mvn = (struct MyVolNode *) mvn->mvn_Node.ln_Succ)
	{
		DeviceList	= (struct DeviceList *) BADDR (DosInfo->di_DevInfo);
		GoOn = TRUE;

		/*-----------------------------------------------------------------*/
		/* Walking the DeviceList once again hunting for DLT_DEVICE types. */
		/*-----------------------------------------------------------------*/
		while (DeviceList && GoOn)
		{
			ptrName = (BYTE *) BADDR (DeviceList->dl_Name);
			len = *ptrName++;

			/*---------------------------------------------------------------*/
			/* Check that device is a volume and that everything else is ok. */
			/*---------------------------------------------------------------*/
			if ((len <= NSIZE)								&&
				 (DeviceList->dl_Type == DLT_DEVICE)	&&
				 (DeviceList->dl_Task))
			{
				UBYTE x [NSIZE];
				strcpy (x, ptrName);
				strcat (x, ":");
				if (VLock = Lock (x, ACCESS_READ))
				{
					if (NameFromLock (VLock, mvn->mvn_DevName, NSIZE))
					{
						if (!(Stricmp (mvn->mvn_Name, mvn->mvn_DevName)))
						{
							strcpy (mvn->mvn_DevName, ptrName);
							strcat (mvn->mvn_DevName, ":");
							GoOn = FALSE;
						}
					}
					UnLock (VLock);
				}
			}
			DeviceList = (struct DeviceList *) BADDR (DeviceList->dl_Next);
		}
	}

	/*-------------------------------------------------------------*/
	/* End this function with sorting the two lists we've created. */
	/*-------------------------------------------------------------*/
	SortList (MyVolList, 0);
	SortList (MyAssList, 0);
}

/**/
/*                            UpdateScrList()									*/
/**/
void UpdateScrList (BOOL Full)
{
	/*----------------------------*/
	/* Remove list from listview. */
	/*----------------------------*/
	GT_SetGadgetAttrs (SLBGads [1], SLBWin, NULL,
							 GTLV_Labels, NULL,
							 TAG_DONE);

	/*--------------------------*/
	/* Find all public screens. */
	/*--------------------------*/
	if (Full)
	{
		FreeAllNodes (MyScrList);
		CollectScrList ();
	}

	/*----------------------------------*/
	/* Resubmit the list to the gadget. */
	/*----------------------------------*/
	GT_SetGadgetAttrs (SLBGads [1], SLBWin, NULL,
							 GTLV_Labels,		MyScrList,
							 GTLV_Selected,	0,
							 TAG_DONE);
}

/**/
/*                            UpdateLibList()									*/
/**/
void UpdateLibList (BOOL Full)
{
	/*----------------------------*/
	/* Remove list from listview. */
	/*----------------------------*/
	GT_SetGadgetAttrs (SLBGads [1], SLBWin, NULL,
							 GTLV_Labels, NULL,
							 TAG_DONE);

	/*---------------------*/
	/* Find all libraries. */
	/*---------------------*/
	if (Full)
	{
		FreeAllNodes (MyLibList);
		CollectLibList ();
	}

	/*----------------------------------*/
	/* Resubmit the list to the gadget. */
	/*----------------------------------*/
	GT_SetGadgetAttrs (SLBGads [1], SLBWin, NULL,
							 GTLV_Labels,		MyLibList,
							 GTLV_Selected,	0,
							 TAG_DONE);
}

/**/
/*                            UpdateDevList()									*/
/**/
void UpdateDevList (BOOL Full)
{
	/*----------------------------*/
	/* Remove list from listview. */
	/*----------------------------*/
	GT_SetGadgetAttrs (SLBGads [1], SLBWin, NULL,
							 GTLV_Labels, NULL,
							 TAG_DONE);

	/*-------------------*/
	/* Find all devices. */
	/*-------------------*/
	if (Full)
	{
		FreeAllNodes (MyDevList);
		CollectDevList ();
	}

	/*----------------------------------*/
	/* Resubmit the list to the gadget. */
	/*----------------------------------*/
	GT_SetGadgetAttrs (SLBGads [1], SLBWin, NULL,
							 GTLV_Labels,		MyDevList,
							 GTLV_Selected,	0,
							 TAG_DONE);
}

/**/
/*                            UpdateTskList()									*/
/**/
void UpdateTskList (BOOL Full)
{
	/*----------------------------*/
	/* Remove list from listview. */
	/*----------------------------*/
	GT_SetGadgetAttrs (SLBGads [1], SLBWin, NULL,
							 GTLV_Labels, NULL,
							 TAG_DONE);

	/*-----------------*/
	/* Find all tasks. */
	/*-----------------*/
	if (Full)
	{
		FreeAllNodes (MyTskList);
		CollectTskList ();
	}

	/*----------------------------------*/
	/* Resubmit the list to the gadget. */
	/*----------------------------------*/
	GT_SetGadgetAttrs (SLBGads [1], SLBWin, NULL,
							 GTLV_Labels,		MyTskList,
							 GTLV_Selected,	0,
							 TAG_DONE);
}

/**/
/*                           UpdatePrtList() 									*/
/**/
void UpdatePrtList (BOOL Full)
{
	/*----------------------------*/
	/* Remove list from listview. */
	/*----------------------------*/
	GT_SetGadgetAttrs (SLBGads [1], SLBWin, NULL,
							 GTLV_Labels, NULL,
							 TAG_DONE);

	/*-----------------*/
	/* Find all tasks. */
	/*-----------------*/
	if (Full)
	{
		FreeAllNodes (MyPrtList);
		CollectPrtList ();
	}

	/*----------------------------------*/
	/* Resubmit the list to the gadget. */
	/*----------------------------------*/
	GT_SetGadgetAttrs (SLBGads [1], SLBWin, NULL,
							 GTLV_Labels,		MyPrtList,
							 GTLV_Selected,	0,
							 TAG_DONE);
}

/**/
/*                            UpdateResList()									*/
/**/
void UpdateResList (BOOL Full)
{
	/*----------------------------*/
	/* Remove list from listview. */
	/*----------------------------*/
	GT_SetGadgetAttrs (SLBGads [1], SLBWin, NULL,
							 GTLV_Labels, NULL,
							 TAG_DONE);

	/*-----------------*/
	/* Find all tasks. */
	/*-----------------*/
	if (Full)
	{
		FreeAllNodes (MyResList);
		CollectResList ();
	}

	/*----------------------------------*/
	/* Resubmit the list to the gadget. */
	/*----------------------------------*/
	GT_SetGadgetAttrs (SLBGads [1], SLBWin, NULL,
							 GTLV_Labels,		MyResList,
							 GTLV_Selected,	0,
							 TAG_DONE);
}

/**/
/*                            UpdateFntList()									*/
/**/
void UpdateFntList (BOOL Full)
{
	/*----------------------------*/
	/* Remove list from listview. */
	/*----------------------------*/
	GT_SetGadgetAttrs (SLBGads [1], SLBWin, NULL,
							 GTLV_Labels, NULL,
							 TAG_DONE);

	/*-----------------*/
	/* Find all tasks. */
	/*-----------------*/
	if (Full)
	{
		if (SLBWin)
		{
			SleepWindow (SLBWin, &SleepReq2);
		}
		FreeAllNodes (MyFntList);
		CollectFntList ();
		if (SLBWin)
		{
			WakeUpWindow (SLBWin, &SleepReq2);
		}
	}

	/*----------------------------------*/
	/* Resubmit the list to the gadget. */
	/*----------------------------------*/
	GT_SetGadgetAttrs (SLBGads [1], SLBWin, NULL,
							 GTLV_Labels,		MyFntList,
							 GTLV_Selected,	0,
							 TAG_DONE);
}

/**/
/*                            UpdateVolList()									*/
/**/
void UpdateVolList (BOOL Full)
{
	/*----------------------------*/
	/* Remove list from listview. */
	/*----------------------------*/
	GT_SetGadgetAttrs (SLBGads [1], SLBWin, NULL,
							 GTLV_Labels, NULL,
							 TAG_DONE);

	/*-----------------*/
	/* Find all tasks. */
	/*-----------------*/
	if (Full)
	{
		if (SLBWin)
		{
			SleepWindow (SLBWin, &SleepReq2);
		}
		FreeAllNodes (MyVolList);
		CollectVolList ();
		if (SLBWin)
		{
			WakeUpWindow (SLBWin, &SleepReq2);
		}
	}

	/*----------------------------------*/
	/* Resubmit the list to the gadget. */
	/*----------------------------------*/
	GT_SetGadgetAttrs (SLBGads [1], SLBWin, NULL,
							 GTLV_Labels,		MyVolList,
							 GTLV_Selected,	0,
							 TAG_DONE);
}

/**/
/*                            UpdateAssList()									*/
/**/
void UpdateAssList (BOOL Full)
{
	/*----------------------------*/
	/* Remove list from listview. */
	/*----------------------------*/
	GT_SetGadgetAttrs (SLBGads [1], SLBWin, NULL,
							 GTLV_Labels, NULL,
							 TAG_DONE);

	/*-----------------*/
	/* Find all tasks. */
	/*-----------------*/
	if (Full)
	{
		if (SLBWin)
		{
			SleepWindow (SLBWin, &SleepReq2);
		}
		FreeAllNodes (MyAssList);
		CollectVolList ();
		if (SLBWin)
		{
			WakeUpWindow (SLBWin, &SleepReq2);
		}
	}

	/*----------------------------------*/
	/* Resubmit the list to the gadget. */
	/*----------------------------------*/
	GT_SetGadgetAttrs (SLBGads [1], SLBWin, NULL,
							 GTLV_Labels,		MyAssList,
							 GTLV_Selected,	0,
							 TAG_DONE);
}

/**/
/*                             CreateANode()										*/
/**/
void CreateANode (struct List *l, struct Node *n)
{
	n->ln_Type	= NT_USER;
	n->ln_Pri	= 0;
	AddTail (l, n);
}

/**/
/*                            FreeAllNodes()										*/
/**/
void FreeAllNodes (struct List *l)
{
	//
	// This function frees all nodes in a list, and deallocates memory for
	// each node. In the end a NewList() is performed on the list, so that
	// it is immidiately ready for use.
	//
	/*-------------*/
	/* L O C A L S */
	/*-------------*/
	struct Node *wn;									// Work node.
	struct Node *nn;									// Next node.

	/*---------*/
	/* C O D E */
	/*---------*/
	wn = l->lh_Head;									// Point to first in list.
	while (nn = wn->ln_Succ)
	{
		if (!mem_pool_ok)
		{
			FreeVec (wn);
		}
		wn = nn;
	}
	NewList (l);
}

/**/
/*                             GetNodeNum()										*/
/**/
struct Node *GetNodeNum (struct List *l, ULONG num)
{
	//
	// This function returns the <num>th node in list <l>, and is called
	// whenever a user clicks on an item in a listview. We don't do error
	// handling here, so be absolutely sure that <num> is valid.
	//
	/*-------------*/
	/* L O C A L S */
	/*-------------*/
	ULONG			count = 0;
	struct Node	*n;

	/*---------*/
	/* C O D E */
	/*---------*/
	/*-------------------------------------------------*/
	/* Traverse down the list, starting from the head. */
	/*-------------------------------------------------*/
	for (n = l->lh_Head; n->ln_Succ; n = n->ln_Succ)
	{
		if (count == num)
		{
			return (n);
		}
		count++;
	}
}
